import torch
import numpy as np

class CachedDataset(torch.utils.data.Dataset):
    def __init__(self, df, features, target_column, transform = None, target_transform = None):
        self.df = df
        self.features = features
        self.target_column = target_column
        self.indices = list(self.df.index)
        self.transform = transform
        self.target_transform = target_transform
    def __len__(self):
        return len(self.indices)
    def __getitem__(self, item):
        index = self.indices[item]
        signal = np.load(self.features.loc[index, 'filename'])
        target = self.df.loc[index, self.target_column]
        if self.transform is not None:
            signal = self.transform(signal)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if isinstance(self.target_column, list):
            target = tuple(zip(target, target.index))
        return signal, target


# dataset.py
import audformat
#import audb
import pandas as pd
import ast
import os
from random import sample
from nkululeko.util import Util
from nkululeko.plots import Plots
import nkululeko.glob_conf as glob_conf
import os.path
from audformat.utils import duration
import nkululeko.filter_data as filter
from nkululeko.filter_data import DataFilter

class Dataset:
    """ Class to represent datasets"""
    name = '' # An identifier for the dataset
    config = None # The configuration 
    db = None # The database object
    df = None # The whole dataframe
    df_train = None # The training split
    df_test = None # The evaluation split

    def __init__(self, name):
        """Constructor setting up name and configuration"""
        self.name = name
        self.target = glob_conf.config['DATA']['target']
        self.util = Util('dataset')
        self.plot = Plots()
        self.limit = int(self.util.config_val_data(self.name, 'limit', 0))
        self.start_fresh = eval(self.util.config_val('DATA', 'no_reuse', 'False'))
        self.is_labeled, self.got_speaker, self.got_gender, self.got_age = False, False, False, False 


    def _get_tables(self):
        tables = []
        targets = self.util.config_val_data(self.name, 'target_tables', False)
        if targets:
            target_tables = ast.literal_eval(targets)
            tables += target_tables
        files = self.util.config_val_data(self.name, 'files_tables', False)
        if files:
            files_tables = ast.literal_eval(files)
            tables += files_tables
        tests = self.util.config_val_data(self.name, 'test_tables', False)
        if tests:
            test_tables = ast.literal_eval(tests)
            tables += test_tables
        trains = self.util.config_val_data(self.name, 'train_tables', False)
        if trains:
            train_tables = ast.literal_eval(trains)
            tables += train_tables
        return tables
    
    def _load_db(self):
        root = self.util.config_val_data(self.name, '', '')
        self.util.debug(f'{self.name}: loading from {root}')
        try:
            self.db = audformat.Database.load(root)
        except FileNotFoundError:
            self.util.error( f'{self.name}: no database found at {root}')
        return root

    def load(self):
        """Load the dataframe with files, speakers and task labels"""
        # store the dataframe
        store = self.util.get_path('store')
        store_file = f'{store}{self.name}.pkl'
        self.util.debug(f'{self.name}: loading ...')
        self.root = self._load_db()
#        self.got_speaker, self.got_gender = False, False 
        if not self.start_fresh and os.path.isfile(store_file):
            self.util.debug(f'{self.name}: reusing previously stored file {store_file}')
            self.df = pd.read_pickle(store_file)
            self.is_labeled = self.target in self.df
            self.got_gender = 'gender' in self.df
            self.got_age = 'age' in self.df
            self.got_speaker = 'speaker' in self.df
            self.util.copy_flags(self, self.df)
            self.util.debug(f'{self.name}: loaded with {self.df.shape[0]} '\
                f'samples: got targets: {self.is_labeled}, got speakers: {self.got_speaker}, '\
                f'got sexes: {self.got_gender}')        
            return
        tables = self._get_tables()
        self.util.debug(f'{self.name}: loading tables: {tables}')
        #db = audb.load(root, )
        # map the audio file paths 
        self.db.map_files(lambda x: os.path.join(self.root, x))
        # the dataframes (potentially more than one) with at least the file names
        df_files = self.util.config_val_data(self.name, 'files_tables', '[\'files\']')
        df_files_tables =  ast.literal_eval(df_files)
        # The label for the target column 
        self.col_label = self.util.config_val_data(self.name, 'label', self.target)
        df, self.is_labeled, self.got_speaker, self.got_gender, self.got_age = self._get_df_for_lists(self.db, df_files_tables)
        if False in {self.is_labeled, self.got_speaker, self.got_gender, self.got_age}:
            try :
            # There might be a separate table with the targets, e.g. emotion or age    
                df_targets = self.util.config_val_data(self.name, 'target_tables', f'[\'{self.target}\']')
                df_target_tables =  ast.literal_eval(df_targets)
                df_target, got_target2, got_speaker2, got_gender2, got_age2 = self._get_df_for_lists(self.db, df_target_tables)
                self.is_labeled = got_target2 or self.is_labeled
                self.got_speaker = got_speaker2 or self.got_speaker
                self.got_gender = got_gender2 or self.got_gender
                self.got_age = got_age2 or self.got_age
                if got_target2:
                    df[self.target] = df_target[self.target]
                if got_speaker2:
                    df['speaker'] = df_target['speaker']
                if got_gender2:
                    df['gender'] = df_target['gender']
                if got_age2:
                    df['age'] = df_target['age']
            except audformat.core.errors.BadKeyError:
                pass

        if self.is_labeled:
            # remember the target in case they get labelencoded later
            df['class_label'] = df[self.target]

        self.df = df
        self.util.debug(f'Loaded database {self.name} with {df.shape[0]} '\
            f'samples: got targets: {self.is_labeled}, got speakers: {self.got_speaker}, '\
            f'got sexes: {self.got_gender}, got age: {self.got_age}')


    def prepare(self):
        # ensure segmented index
        self.df = self.util.make_segmented_index(self.df)
        self.util.copy_flags(self, self.df)
        # add duration
        if 'duration' not in self.df:
            start = self.df.index.get_level_values(1)
            end = self.df.index.get_level_values(2)
            self.df['duration']=(end-start).total_seconds()

        # Perform some filtering if desired
        required = eval(self.util.config_val_data(self.name, 'required', 'False'))
        if required:
            pre = self.df.shape[0]
            self.df = self.df[self.df[required].notna()]
            post = self.df.shape[0]
            self.util.debug(f'{self.name}: kept {post} samples with {required} (from {pre}, filtered {pre-post})')

        datafilter = DataFilter(self.df)
        self.df = datafilter.all_filters(data_name=self.name)

        self.util.debug(f'{self.name}: loaded data with {self.df.shape[0]} '\
            f'samples: got targets: {self.is_labeled}, got speakers: {self.got_speaker}, '\
            f'got sexes: {self.got_gender}')

        if self.got_speaker and self.util.config_val_data(self.name, 'rename_speakers', False):
            # we might need to append the database name to all speakers in case other datbaases have the same speaker names
            self.df.speaker = self.df.speaker.apply(lambda x: self.name+x)

        # store the dataframe
        store = self.util.get_path('store')
        store_file = f'{store}{self.name}.pkl'
        self.df.to_pickle(store_file)


    def _get_df_for_lists(self, db, df_files):
        is_labeled, got_speaker, got_gender, got_age = False, False, False, False
        df = pd.DataFrame()
        for table in df_files:
            source_df = db.tables[table].df
            # create a dataframe with the index (the filenames)
            df_local = pd.DataFrame(index=source_df.index)
            # try to get the targets from this dataframe
            try:
                # try to get the target values
                df_local[self.target] = source_df[self.col_label]
                is_labeled = True
            except (KeyError, ValueError, audformat.errors.BadKeyError) as e:
                pass
            try:
                # try to get the speaker values
                df_local['speaker'] = source_df['speaker']
                got_speaker = True
            except (KeyError, ValueError, audformat.errors.BadKeyError) as e:
                pass
            try:
                # try to get the gender values
                df_local['gender'] = source_df['gender']
                got_gender = True
            except (KeyError, ValueError, audformat.errors.BadKeyError) as e:
                pass
            try:
                # try to get the age values
                df_local['age'] = source_df['age']
                got_age = True
            except (KeyError, ValueError, audformat.errors.BadKeyError) as e:
                pass
            try:
                # also it might be possible that the sex is part of the speaker description
                df_local['gender'] = db[table]['speaker'].get(map='gender')
                got_gender = True
            except (ValueError, audformat.errors.BadKeyError) as e:
                pass
            try:
                # also it might be possible that the age is part of the speaker description
                df_local['age'] = db[table]['speaker'].get(map='age')
                got_age = True
            except (ValueError, audformat.errors.BadKeyError) as e:
                pass
            try:
                # same for the target, e.g. "age"
                df_local[self.target] = db[table]['speaker'].get(map=self.target)
                is_labeled = True
            except (ValueError, audformat.core.errors.BadKeyError) as e:
                pass
            df = pd.concat([df, df_local])
        return df, is_labeled, got_speaker, got_gender, got_age


    def split(self):
        """Split the datbase into train and development set"""
        store = self.util.get_path('store')
        storage_test = f'{store}{self.name}_testdf.pkl'
        storage_train = f'{store}{self.name}_traindf.pkl'
        split_strategy = self.util.config_val_data(self.name,'split_strategy', 'database')
        self.util.debug(f'splitting database {self.name} with strategy {split_strategy}')
        # 'database' (default), 'speaker_split', 'specified', 'reuse'
        if split_strategy != 'speaker_split' and not self.start_fresh:
            # check if the splits have been computed previously (not for speaker split)
            if os.path.isfile(storage_train) and os.path.isfile(storage_test):
                # if self.util.config_val_data(self.name, 'test_tables', False):
                self.util.debug(f'splits: reusing previously stored test file {storage_test}')
                self.df_test = pd.read_pickle(storage_test)
                self.util.debug(f'splits: reusing previously stored train file {storage_train}')
                self.df_train = pd.read_pickle(storage_train)

                return
            elif os.path.isfile(storage_train):
                self.util.debug(f'splits: reusing previously stored train file {storage_train}')
                self.df_train = pd.read_pickle(storage_train)
                self.df_test = pd.DataFrame()
                return
            elif os.path.isfile(storage_test):
                self.util.debug(f'splits: reusing previously stored test file {storage_test}')
                self.df_test = pd.read_pickle(storage_test)
                self.df_train = pd.DataFrame()
                return
        if split_strategy == 'database':
            #  use the splits from the database
            testdf = self.db.tables[self.target+'.test'].df
            traindf = self.db.tables[self.target+'.train'].df
            # use only the train and test samples that were not perhaps filtered out by an earlier processing step
            self.df_test = self.df.loc[self.df.index.intersection(testdf.index)]
            self.df_train = self.df.loc[self.df.index.intersection(traindf.index)]
        elif split_strategy == 'train':
            self.df_train = self.df
            self.df_test = pd.DataFrame()
        elif split_strategy == 'test':
            self.df_test = self.df
            self.df_train = pd.DataFrame()
        elif split_strategy == 'specified':
            traindf, testdf = pd.DataFrame(), pd.DataFrame()
            # try to load some dataframes for testing
            entry_test_tables = self.util.config_val_data(self.name, 'test_tables', False)
            if entry_test_tables: 
                test_tables =  ast.literal_eval(entry_test_tables)
                for test_table in test_tables:
                    testdf = pd.concat([testdf, self.db.tables[test_table].df]) 
            entry_train_tables = self.util.config_val_data(self.name, 'train_tables', False)
            if entry_train_tables: 
                train_tables =  ast.literal_eval(entry_train_tables)
                for train_table in train_tables:
                    traindf = pd.concat([traindf, self.db.tables[train_table].df])
            # use only the train and test samples that were not perhaps filtered out by an earlier processing step
            #testdf.index.map(lambda x: os.path.join(self.root, x))
#            testdf.index = testdf.index.to_series().apply(lambda x: self.root+x)
            testdf = testdf.set_index(audformat.utils.to_segmented_index(testdf.index, allow_nat=False))
            traindf = traindf.set_index(audformat.utils.to_segmented_index(traindf.index, allow_nat=False))
            self.df_test = self.df.loc[self.df.index.intersection(testdf.index)]
            self.df_train = self.df.loc[self.df.index.intersection(traindf.index)]
            # it might be necessary to copy the target values 
            try:
                self.df_test[self.target] = testdf[self.target]
            except KeyError:
                pass # if the dataframe is empty
            try:
                self.df_train[self.target] = traindf[self.target]
            except KeyError:
                pass # if the dataframe is empty
        elif split_strategy == 'speaker_split':
            self.split_speakers()
        elif split_strategy == 'random':
            self.random_split()
        elif split_strategy == 'reuse':
            self.util.debug(f'{self.name}: trying to reuse data splits')
            self.df_test = pd.read_pickle(storage_test)
            self.df_train = pd.read_pickle(storage_train)

        if self.df_test.shape[0]>0:
            self.df_test = self.finish_up(self.df_test, storage_test)
        if self.df_train.shape[0]>0:
            self.df_train = self.finish_up(self.df_train, storage_train)

        self.util.debug(f'{self.name}: {self.df_test.shape[0]} samples in test and {self.df_train.shape[0]} samples in train')

        
    def finish_up(self, df, storage):
        # Bin target values if they are continuous but a classification experiment should be done
        # self.check_continuous_classification(df)
        # remember the splits for future use
        df.is_labeled = self.is_labeled
        self.df_test.is_labeled = self.is_labeled
        df.to_pickle(storage)
        return df

    def split_speakers(self):
        """One way to split train and eval sets: Specify percentage of evaluation speakers"""
        test_percent = int(self.util.config_val_data(self.name, 'testsplit', 50))
        df = self.df
        s_num = df.speaker.nunique()
        test_num = int(s_num * (test_percent/100))        
        test_spkrs =  sample(list(df.speaker.unique()), test_num)
        self.df_test = df[df.speaker.isin(test_spkrs)]
        self.df_train = df[~df.index.isin(self.df_test.index)]
        self.util.debug(f'{self.name}: [{self.df_train.shape[0]}/{self.df_test.shape[0]}] samples in train/test')
        # because this generates new train/test sample quantaties, the feature extraction has to be done again
        glob_conf.config['FEATS']['needs_feature_extraction'] = 'True'

    def random_split(self):
        """One way to split train and eval sets: Specify percentage of random samples"""
        test_percent = int(self.util.config_val_data(self.name, 'testsplit', 50))
        df = self.df
        s_num = len(df)
        test_num = int(s_num * (test_percent/100))        
        test_smpls =  sample(list(df.index), test_num)
        self.df_test = df[df.index.isin(test_smpls)]
        self.df_train = df[~df.index.isin(self.df_test.index)]
        self.util.debug(f'{self.name}: [{self.df_train.shape[0]}/{self.df_test.shape[0]}] samples in train/test')
        # because this generates new train/test sample quantaties, the feature extraction has to be done again
        glob_conf.config['FEATS']['needs_feature_extraction'] = 'True'

    def _add_labels(self, df):
        df.is_labeled = self.is_labeled
        df.got_gender = self.got_gender
        df.got_age = self.got_age
        df.got_speaker = self.got_speaker
        return df

    def prepare_labels(self):
        strategy = self.util.config_val('DATA', 'strategy', 'train_test')
        only_tests = eval(self.util.config_val('DATA', 'tests', 'False'))
        if strategy == 'cross_data' or only_tests:
            self.df = self.map_labels(self.df)
            # Bin target values if they are continuous but a classification experiment should be done
            self.map_continuous_classification(self.df)
            self.df = self._add_labels(self.df)
            if self.util.config_val_data(self.name, 'value_counts', False):
                if not self.got_gender or not self.got_speaker:
                    self.util.error('can\'t plot value counts if no speaker or gender is given')
                else:
                    self.plot.describe_df(self.name, self.df, self.target, f'{self.name}_distplot')
        elif strategy == 'train_test':        
            self.df_train = self.map_labels(self.df_train)
            self.df_test = self.map_labels(self.df_test)
            self.map_continuous_classification(self.df_train)
            self.map_continuous_classification(self.df_test)
            self.df_train = self._add_labels(self.df_train)
            self.df_test = self._add_labels(self.df_test)
            if self.util.config_val_data(self.name, 'value_counts', False):
                if not self.got_gender or not self.got_speaker:
                    self.util.error('can\'t plot value counts if no speaker or gender is given')
                else:
                    self.plot.describe_df(self.name, self.df_train, self.target, f'{self.name}_train_distplot')
                    self.plot.describe_df(self.name, self.df_test, self.target, f'{self.name}_test_distplot')


    def map_labels(self, df):
        pd.options.mode.chained_assignment = None
        if df.shape[0]==0 or not self.util.exp_is_classification() \
            or self.check_continuous_classification():
            return df
        """Rename the labels and remove the ones that are not needed."""
        target = glob_conf.config['DATA']['target']
        # see if a special mapping should be used
        mappings = self.util.config_val_data(self.name, 'mapping', False)
        if mappings:        
            mapping = ast.literal_eval(mappings)
            df[target] = df[target].map(mapping)
            self.util.debug(f'{self.name}: mapped {mapping}')
        # remove labels that are not in the labels list
        labels = self.util.config_val('DATA', 'labels', False)
        if labels:
            labels = ast.literal_eval(labels)
            df = df[df[target].isin(labels)]
        else:
            labels = df[target].unique().codes
        # try:
        # except KeyError:
        #     pass
        # remember in case they get encoded later
        df['class_label'] = df[target]
        return df

    def check_continuous_classification(self):
        datatype = self.util.config_val('DATA', 'type', 'False')
        if self.util.exp_is_classification() and datatype == 'continuous':
            return True
        return False

    def map_continuous_classification(self, df):
        """Map labels to bins for continuous data that should be classified"""
        if self.check_continuous_classification():
            self.util.debug(f'{self.name}: binning continuous variable to categories')
            cat_vals = self.util.continuous_to_categorical(df[self.target])
            df[self.target] = cat_vals
            labels = ast.literal_eval(glob_conf.config['DATA']['labels'])
            df['class_label'] = df[self.target]
#            print(df['class_label'].unique())
            for i, l in enumerate(labels):
                df['class_label'] = df['class_label'].replace(i, str(l))


# feats_audmodel.py
from nkululeko.featureset import Featureset
import os
import pandas as pd
import audeer
import nkululeko.glob_conf as glob_conf
import audonnx
import numpy as np
import audinterface

class AudModelSet(Featureset):
    """
        Embeddings from the wav2vec2. based model finetuned on MSPPodcast emotions, described in the paper
        "Dawn of the transformer era in speech emotion recognition: closing the valence gap"
        https://arxiv.org/abs/2203.07378
    """
    def __init__(self, name, data_df):
        super().__init__(name, data_df)
        model_url = 'https://zenodo.org/record/6221127/files/w2v2-L-robust-12.6bc4a7fd-1.1.0.zip'
        model_root = self.util.config_val('FEATS', 'aud.model', './audmodel/')
        if not os.path.isdir(model_root):
            cache_root = audeer.mkdir('cache')
            model_root = audeer.mkdir(model_root)
            archive_path = audeer.download_url(model_url, cache_root, verbose=True)
            audeer.extract_archive(archive_path, model_root)
        device = self.util.config_val('MODEL', 'device', 'cpu')
        self.model = audonnx.load(model_root, device=device)


    def extract(self):
        """Extract the features based on the initialized dataset or re-open them when found on disk."""
        store = self.util.get_path('store')
        store_format = self.util.config_val('FEATS', 'store_format', 'pkl')
        storage = f'{store}{self.name}.{store_format}'
        extract = eval(self.util.config_val('FEATS', 'needs_feature_extraction', 'False'))
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if no_reuse or extract or not os.path.isfile(storage):
            self.util.debug('extracting audmodel features, this might take a while...')
            hidden_states = audinterface.Feature(
                self.model.labels('hidden_states'),
                process_func=self.model,
                process_func_args={
                    'outputs': 'hidden_states',
                },
                sampling_rate=16000,    
                resample=True,    
                num_workers=5,
                verbose=True,
            )
            self.df = hidden_states.process_index(self.data_df.index)
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'False'
            except KeyError:
                pass
        else:
            self.util.debug('reusing extracted audmodel features.')
            self.df = self.util.get_store(storage, store_format)


    def extract_sample(self, signal, sr):
        result = self.model(signal, sr)
        return np.asarray(result['hidden_states'].flatten())


# result.py

class Result:
    def __init__(self, test, train, loss, measure):
        self.test = test
        self.train = train
        self.loss = loss 
        self.measure = measure

    def get_test_result(self):
        return f'test: {self.test:.3f} {self.measure}'

    def to_string(self):
        return f'test: {self.test} {self.measure}, train: {self.train} {self.measure}, loss: {self.loss}'


# mld_fset.py
from nkululeko.featureset import Featureset
import sys
import os
import pandas as pd
import numpy as np
from nkululeko.util import Util 
import nkululeko.glob_conf as glob_conf

class MLD_set(Featureset):

    def __init__(self, name, data_df):
        super().__init__(name, data_df)   
        mld_path = self.util.config_val('FEATS', 'mld.model', None)
        sys.path.append(mld_path)

    def extract(self):
        store = self.util.get_path('store')
        storage = f'{store}{self.name}.pkl'
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if no_reuse:
            os.remove(storage)
        if not os.path.isfile(storage):
            self.util.debug('extracting midleveldescriptor features, this might take a while...')
        else:
            self.util.debug('reusing previously extracted midleveldescriptor features')
        import midlevel_descriptors as mld
        fex_mld = mld.MLD()
        self.df = fex_mld.extract_from_index(index=self.data_df, cache_path=storage)
        self.util.debug(f'MLD feats shape: {self.df.shape}')
        # shouldn't happen
        # replace NANa with column means values
        self.util.debug('MLD extractor: checking for NANs...')
        for i, col in enumerate(self.df.columns):
            if self.df[col].isnull().values.any():
                self.util.debug(f'{col} includes {self.df[col].isnull().sum()} nan, inserting mean values')
                self.df[col] = self.df[col].fillna(self.df[col].mean())

        try:
            # use only samples that have a minimum number of syllables
            min_syls = int(glob_conf.config['FEATS']['min_syls'])
            self.df = self.df[self.df['hld_nSyl']>=min_syls]
        except KeyError:
            pass
        if self.df.isna().to_numpy().any():
            self.util.error('feats 0: NANs exist')
        self.df = self.df.astype(float)



# dataset.py
from nkululeko.dataset import Dataset

class Ravdess(Dataset):
    """Class to represent the Berlin EmoDB"""
    name = 'ravdess' # The name    

    def __init__(self):
        """Constructor setting the name"""
        Dataset.__init__(self, self.name) 

    def load(self):
        Dataset.load(self) 
        df = self.df
        prev = df.shape[0]
        df = df[~df.index.str.contains('song')]
        now = df.shape[0]
        self.util.debug(f'removed {prev-now} songs from ravdess dataframe')
        self.df = df



# model_tree.py

from sklearn.tree import DecisionTreeClassifier
from nkululeko.model import Model

class Tree_model(Model):
    """An Tree model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.clf = DecisionTreeClassifier() # set up the classifier




# feats_trill.py
from numpy.core.numeric import tensordot
from nkululeko.featureset import Featureset
import pandas as pd
from nkululeko.util import Util 
import nkululeko.glob_conf as glob_conf
import audiofile as af
import os
import tensorflow as tf
# Import TF 2.X and make sure we're running eager.
assert tf.executing_eagerly()
import tensorflow_hub as hub

class TRILLset(Featureset):
    """A feature extractor for the Google TRILL embeddings"""
    """https://ai.googleblog.com/2020/06/improving-speech-representations-and.html"""

    # Initialization of the class
    def __init__(self, name, data_df):
        """
        Initialize the class with name, data and Util instance
        Also loads the model from hub
        
        :param name: Name of the class
        :type name: str
        :param data_df: Data of the class
        :type data_df: DataFrame
        :return: None
        """
        super().__init__(name, data_df)   
        # Load the model from the configured path
        model_path = self.util.config_val('FEATS', 'trill.model', \
            'https://tfhub.dev/google/nonsemantic-speech-benchmark/trill/3')
        self.module = hub.load(model_path)


    def extract(self):
        store = self.util.get_path('store')
        storage = f'{store}{self.name}.pkl'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug('extracting TRILL embeddings, this might take a while...')
            emb_series = pd.Series(index = self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, file in enumerate(self.data_df.index.get_level_values(0)):
                emb = self.getEmbeddings(file)
                emb_series[idx] = emb
                if idx%10==0:
                    self.util.debug(f'TRILL: {idx} of {length} done')
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index) 
            self.df.to_pickle(storage)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'false'
            except KeyError:
                pass
        else:
            self.util.debug('reusing extracted TRILL embeddings')
            self.df = pd.read_pickle(storage)

    def embed_wav(self, wav): 
        if len(wav.shape) > 1:
            wav = tf.reduce_mean(wav, axis=0)
            
        emb_dict = self.module(samples=wav, sample_rate=tf.constant(16000))
        return emb_dict['embedding']

    def getEmbeddings(self, file):
        wav = af.read(file)[0]
        wav = tf.convert_to_tensor(wav)
        emb_short = self.embed_wav(wav)
        # you get one embedding per frame, we use the mean for all the frames
        emb_short = emb_short.numpy().mean(axis=0)
        return emb_short



# feats_wav2vec2.py

from nkululeko.util import Util
from nkululeko.featureset import Featureset
import os
import pandas as pd
import os
import nkululeko.glob_conf as glob_conf
import transformers
from transformers.models.wav2vec2.modeling_wav2vec2 import Wav2Vec2Model
import torch

import audiofile

class Wav2vec2(Featureset):
    """Class to extract wav2vec2 embeddings (https://huggingface.co/facebook/wav2vec2-large-robust-ft-swbd-300h)"""

    def __init__(self, name, data_df):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df)
        self.device = self.util.config_val('MODEL', 'device', 'cpu')
        self.model_initialized = False



    def init_model(self):
        # load model
        self.util.debug('loading wav2vec model...')
        model_path = self.util.config_val('FEATS', 'wav2vec.model', 'wav2vec2-large-robust-ft-swbd-300h')
        self.processor = transformers.Wav2Vec2Processor.from_pretrained(model_path)
        self.model = Wav2Vec2Model.from_pretrained(model_path).to(self.device)
        print(f'intialized vec model on {self.device}')
        self.model.eval()
        self.model_initialized = True


    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path('store')
        storage = f'{store}{self.name}.pkl'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug('extracting wav2vec2 embeddings, this might take a while...')
            emb_series = pd.Series(index = self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(self.data_df.index.to_list()):
                signal, sampling_rate = audiofile.read(file, offset=start.total_seconds(), duration=(end-start).total_seconds(), always_2d=True)
                #signal, sampling_rate = audiofile.read(audio_path, always_2d=True)
                emb = self.get_embeddings(signal, sampling_rate)
                emb_series[idx] = emb
                if idx%10==0:
                    self.util.debug(f'Wav2vec2: {idx} of {length} done')
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index) 
            self.df.to_pickle(storage)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'false'
            except KeyError:
                pass
        else:
            self.util.debug('reusing extracted wav2vec2 embeddings')
            self.df = pd.read_pickle(storage)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                print(nanrows)
                self.util.error(f'got nan: {self.df.shape} {self.df.isnull().sum().sum()}')


    def get_embeddings(self, signal, sampling_rate):
        r"""Extract embeddings from raw audio signal."""
        with torch.no_grad():
            # run through processor to normalize signal
            # always returns a batch, so we just get the first entry
            # then we put it on the device
            y = self.processor(signal, sampling_rate=sampling_rate)
            y = y['input_values'][0]
            y = torch.from_numpy(y).to(self.device)
        
            # run through model
            # first entry contains hidden state
            y = self.model(y)[0]
        
            # pool result and convert to numpy
            y = torch.mean(y, dim=1)
            y = y.detach().cpu().numpy()
    
        return y.flatten()

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr)
        return feats


# model_gmm.py

from sklearn import mixture
from nkululeko.model import Model

class GMM_model(Model):
    """An GMM model"""

    is_classifier = True
    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        n_components = int(self.util.config_val('MODEL', 'GMM_components', '4'))         
        covariance_type = self.util.config_val('MODEL', 'GMM_covariance_type', 'full') 
        self.clf = mixture.GaussianMixture(n_components=n_components, covariance_type=covariance_type)
        # set up the classifier




# feats_praat.py
from nkululeko.featureset import Featureset
import os
import pandas as pd
import nkululeko.glob_conf as glob_conf
from nkululeko import feinberg_praat
import ast

class Praatset(Featureset):
    """
    a feature extractor for the Praat software, based on 
    David R. Feinberg's Praat scripts for the parselmouth python interface.
    https://osf.io/6dwr3/

    """
    def __init__(self, name, data_df):
        super().__init__(name, data_df)

    def extract(self):
        """Extract the features based on the initialized dataset or re-open them when found on disk."""
        store = self.util.get_path('store')
        store_format = self.util.config_val('FEATS', 'store_format', 'pkl')
        storage = f'{store}{self.name}.{store_format}'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug('extracting Praat features, this might take a while...')
            self.df = feinberg_praat.compute_features(self.data_df.index)
            self.df = self.df.set_index(self.data_df.index)
            for i, col in enumerate(self.df.columns):
                if self.df[col].isnull().values.any():
                    self.util.debug(f'{col} includes {self.df[col].isnull().sum()} nan, inserting mean values')
                    self.df[col] = self.df[col].fillna(self.df[col].mean())

            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'false'
            except KeyError:
                pass
        else:
            self.util.debug(f'reusing extracted Praat features: {storage}.')
            self.df = self.util.get_store(storage, store_format)
        self.util.debug(f'praat feature names: {self.df.columns}')
        self.df = self.df.astype(float)



    def extract_sample(self, signal, sr):
        self.util.error('feats_praat: extracting single samples not implemented yet')
        feats = None
        return feats
    
    def filter(self):
        # use only the features that are indexed in the target dataframes
        self.df = self.df[self.df.index.isin(self.data_df.index)]
        try: 
            # use only some features
            selected_features = ast.literal_eval(glob_conf.config['FEATS']['praat.features'])
            self.util.debug(f'selecting features from Praat: {selected_features}')
            sel_feats_df = pd.DataFrame()
            hit = False
            for feat in selected_features:
                try:
                    sel_feats_df[feat] = self.df[feat]
                    hit = True
                except KeyError:
                    pass
            if hit:
                self.df = sel_feats_df
                self.util.debug(f'new feats shape after selecting Praat features: {self.df.shape}')
        except KeyError:
            pass



# model_knn_reg.py

from sklearn.neighbors import KNeighborsRegressor
from nkululeko.model import Model

class KNN_reg_model(Model):
    """An KNN model"""

    is_classifier = False

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        method = self.util.config_val('MODEL', 'KNN_weights', 'uniform') 
        k = int(self.util.config_val('MODEL', 'K_val', '5'))
        self.clf = KNeighborsRegressor(n_neighbors=k, weights=method) # set up the classifier




# model_svm.py

from sklearn import svm
from nkululeko.model import Model

class SVM_model(Model):
    """An SVM model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        c = float(self.util.config_val('MODEL', 'C_val', '0.001'))
        self.clf = svm.SVC(kernel='linear', C=c, gamma='auto') # set up the classifier

    def set_C(self, c):
        """Set the C parameter"""
        self.clf.C = c



# feats_analyser.py
import pandas as pd
from nkululeko.util import Util 
from nkululeko.plots import Plots
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeRegressor
import matplotlib.pyplot as plt

class FeatureAnalyser:
 

    def __init__(self, label, df_labels, df_features):
        self.util = Util('feats_analyser')
        target = self.util.config_val('DATA', 'target', 'emotion')    
        self.y = df_labels[target]
        self.df_labels = df_labels
        self.X = df_features
        self.label = label


    def analyse(self):
        model_s = self.util.config_val('EXPL', 'model', 'log_reg')
        max_feat_num = int(self.util.config_val('EXPL', 'max_feats', '10'))
        importance = None
        self.util.debug('analysing features...')
        if self.util.exp_is_classification():
            if model_s == 'log_reg':
                model = LogisticRegression()
                model.fit(self.X, self.y)
                importance = model.coef_[0]
            elif model_s == 'tree':
                model = DecisionTreeClassifier()
                model.fit(self.X, self.y)
                importance = model.feature_importances_
                plot_tree = eval(self.util.config_val('EXPL', 'plot_tree', 'False'))
                if plot_tree:
                    plots = Plots()
                    plots.plot_tree(model, self.X)
            else:
                self.util.error(f'invalid analysis method: {model}')
        else: # regression experiment
            if model_s == 'lin_reg':
                model = LinearRegression()
                model.fit(self.X, self.y)
                importance = model.coef_
            elif model_s == 'tree':
                model = DecisionTreeRegressor()
                model.fit(self.X, self.y)
                importance = model.feature_importances_
            else:
                self.util.error(f'invalid analysis method: {model_s}')

        df_imp = pd.DataFrame({'feats':self.X.columns, 'importance':importance})
        df_imp = df_imp.sort_values(by='importance', ascending=False).iloc[:max_feat_num]
        ax = df_imp.plot(x='feats', y='importance', kind='bar')
        ax.set(title=f'{self.label} samples')
        plt.tight_layout()
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val('PLOT', 'format', 'png')
        filename = f'{fig_dir}{exp_name}EXPL_{model_s}.{format}'
        plt.savefig(filename)
        fig = ax.figure
        fig.clear()
        plt.close(fig)
        # result file
        res_dir = self.util.get_path('res_dir')
        file_name = f'{res_dir}{self.util.get_exp_name(only_data=True)}EXPL_{model_s}.txt'
        with open(file_name, "w") as text_file:
            text_file.write(f'features in order of decreasing importance according to model {model_s}:\n'+
            f'{str(df_imp.feats.values)}\n')

        df_imp.to_csv(file_name, mode='a')

                # check if feature distributions should be plotted
        plot_feats = self.util.config_val('EXPL', 'feature_distributions', False)
        if plot_feats: 
            sample_selection = self.util.config_val('EXPL', 'sample_selection', 'all')
            if self.util.exp_is_classification():
                for feature in df_imp.feats:
                    # plot_feature(self, title, feature, label, df_labels, df_features):
                    _plots = Plots()
                    _plots.plot_feature(sample_selection, feature, 'class_label', self.df_labels, self.X)
            else:
                self.util.debug('can\'t plot feature distributions if not classification')



# util.py
import audeer
import ast 
import sys
import nkululeko.glob_conf as glob_conf
import numpy as np
import os.path
import configparser
import audformat
import pandas as pd

class Util:

    stopvals = [False, 'False', 'classification', 'png']

    def __init__(self, caller=None):
        self.got_data_roots = self.config_val('DATA', 'root_folders', False)
        if self.got_data_roots:
            # if there is a global data rootfolder file, read from there
            if not os.path.isfile(self.got_data_roots):
                self.error(f'no such file: {self.got_data_roots}')
            self.data_roots = configparser.ConfigParser()
            self.data_roots.read(self.got_data_roots)
        if caller is not None:
            self.caller = caller
        else:  
            self.caller = ''

    def get_path(self, entry):
        """
        This method allows the user to get the directory path for the given argument.
        """
        root = glob_conf.config['EXP']['root']
        name = glob_conf.config['EXP']['name']
        try:
            entryn = glob_conf.config['EXP'][entry]
        except KeyError:
            # some default values
            if entry == 'fig_dir':
                entryn = './images/'
            elif entry == 'res_dir':
                entryn = './results/'
            elif entry == 'model_dir':
                entryn = './models/'
            else:
                entryn = './store/'

        # Expand image, model and result directories with run index
        if entry == 'fig_dir' or entry == 'res_dir' or entry == 'model_dir':
            run = self.config_val('EXP', 'run', 0)
            entryn = entryn +  f'run_{run}/'

        dir_name = f'{root}{name}/{entryn}'
        audeer.mkdir(dir_name)
        return dir_name
    

    def config_val_data(self, dataset, key, default):
        """ 
        Retrieve a configuration value for datasets.
        If the value is present in the experiment configuration it will be used, else 
        we look in a global file specified by the root_folders value.
        """
        configuration = glob_conf.config
        try:
            if len(key)>0:
                return configuration['DATA'][dataset+'.'+key]
            else:
                return configuration['DATA'][dataset]
        except KeyError:
            if self.got_data_roots:
                try:
                    if len(key)>0:
                        return self.data_roots['DATA'][dataset+'.'+key]
                    else:
                        return self.data_roots['DATA'][dataset]
                except KeyError:
                    if not default in self.stopvals:
                        self.debug(f'value for {key} not found, using default: {default}')
                    return default
            if not default in self.stopvals:
                self.debug(f'value for {key} not found, using default: {default}')
            return default



    def get_save_name(self):
        """Return a relative path to a name to save the experiment"""
        store = self.get_path('store')
        return f'{store}/{self.get_exp_name()}.pkl'

    def is_categorical(self, pd_series):
        """Check if a dataframe column is categorical"""
        return pd_series.dtype.name == 'object'


    def get_res_dir(self):
        root = glob_conf.config['EXP']['root']
        name = glob_conf.config['EXP']['name']
        dir_name = f'{root}{name}/results/'
        audeer.mkdir(dir_name)
        return dir_name

    def make_segmented_index(self, df):
        if len(df)==0:
            return df
        if not isinstance(df.index, pd.MultiIndex):
            df.index = audformat.utils.to_segmented_index(df.index, allow_nat=False)
        return df

    def _get_value_descript(self, section, name):
        if self.config_val(section, name, False):
            val = self.config_val(section, name, False)
            return f'_{name}-{str(val)}'
        return ''

    def get_exp_name(self, only_train = False, only_data = False):
        if only_train:
            # try to get only the train tables
            trains_val = self.config_val('DATA', 'trains', False)
            if trains_val:
                ds = '_'.join(ast.literal_eval(glob_conf.config['DATA']['trains']))
            else:
                # else use all the data
                ds = '_'.join(ast.literal_eval(glob_conf.config['DATA']['databases']))
        else:
            ds = '_'.join(ast.literal_eval(glob_conf.config['DATA']['databases']))
        mt = ''
        if not only_data:
            mt = f'_{glob_conf.config["MODEL"]["type"]}'
        ft = '_'.join(ast.literal_eval(glob_conf.config['FEATS']['type']))
        ft += '_'
        set = self.config_val('FEATS', 'set', False)
        set_string = ''
        if set:
            set_string += set
        # lr_string = ''
        # if self.config_val('MODEL', 'learning_rate', False) and not only_data:
        #     lr = self.config_val('MODEL', 'learning_rate', False)
        #     lr_string = f'_lr-{str(lr)}'
        # logo_string = ''
        # if self.config_val('MODEL', 'logo', False):
        #     logo = self.config_val('MODEL', 'logo', False)
        #     logo_string = f'_logo-{str(logo)}'
        # loss_string = ''
        # if self.config_val('MODEL', 'loss', False) and not only_data:
        #     loss = self.config_val('MODEL', 'loss', False)
        #     loss_string = f'_loss-{loss}'
        # drop_string = ''
        # if self.config_val('MODEL', 'drop', False) and not only_data:
        #     drop = self.config_val('MODEL', 'drop', False)
        #     drop_string = f'_drop-{str(drop)}'
        layer_string = ''
        layer_s = self.config_val('MODEL', 'layers', False)
        if layer_s and not only_data:
            layers = ast.literal_eval(layer_s)
            sorted_layers = sorted(layers.items(), key=lambda x: x[1])
            for l in sorted_layers:
                layer_string += f'{str(l[1])}-'
        return_string = f'{ds}{mt}_{ft}{set_string}'\
            f'{layer_string[:-1]}'
        options = [['MODEL', 'drop'], ['MODEL', 'loss'], ['MODEL', 'logo'], 
                   ['MODEL', 'learning_rate'],  ['MODEL', 'k_fold_cross']]
        for option in options:
            return_string += self._get_value_descript(option[0], option[1])
        # return_string = f'{ds}{mt}_{ft}{set_string}'\
        #     f'{layer_string[:-1]}{lr_string}{drop_string}{loss_string}{logo_string}'.replace('__','')
        return return_string.replace('__','')

    def get_plot_name(self):
        try:
            plot_name = glob_conf.config['PLOT']['name']
        except KeyError:
            plot_name = self.get_exp_name()
        return plot_name

    def exp_is_classification(self):
        type = self.config_val('EXP', 'type', 'classification')
        if type=='classification':
            return True
        return False

    def error(self, message):
        print(f'ERROR {self.caller}: {message}')
        sys.exit()

    def warn(self, message):
        print(f'WARNING {self.caller}: {message}')

    def debug(self, message):
        print(f'DEBUG {self.caller}: {message}')

    def set_config_val(self, section, key, value):
        try:
            # does the section already exists?
            glob_conf.config[section][key] = str(value)
        except KeyError:
            glob_conf.config.add_section(section)
            glob_conf.config[section][key] = str(value)
    
    def check_df(self, i, df):
        """Check a dataframe"""
        print(f'check {i}: {df.shape}')
        print(df.head(1)
        )
    def config_val(self, section, key, default):
        try:
            return glob_conf.config[section][key]
        except KeyError:
            if not default in self.stopvals:
                self.debug(f'value for {key} not found, using default: {default}')
            return default
            
    def config_val_list(self, section, key, default):
        try:
            return ast.literal_eval(glob_conf.config[section][key])
        except KeyError:
            if not default in self.stopvals:
                self.debug(f'value for {key} not found, using default: {default}')
            return default
            
    def get_labels(self):
        # try:
        #     labels = glob_conf.label_encoder.classes_
        # except AttributeError:
        labels = ast.literal_eval(glob_conf.config['DATA']['labels'])
        return labels

    def continuous_to_categorical(self, array):
        bins = ast.literal_eval(glob_conf.config['DATA']['bins'])
        result =  np.digitize(array, bins)-1
        return result

    def print_best_results(self, best_reports):
        res_dir = self.get_res_dir()
        # go one level up above the "run" level
        all = ''
        vals = np.empty(0)
        for report in best_reports:
            all += str(report.result.test) + ', '
            vals = np.append(vals, report.result.test)
        file_name = f'{res_dir}{self.get_exp_name()}_runs.txt'
        with open(file_name, "w") as text_file:
            text_file.write(all)
            text_file.write(f'\nmean: {vals.mean():.3f}, max: {vals.max():.3f}, max_index: {vals.argmax()}')

    def write_store(self, df, storage, format):
        if format == 'pkl':
            df.to_pickle(storage)
        elif format == 'csv':
            df.to_csv(storage)
        else:
            self.error(f'unkown store format: {format}')
            
    def get_store(self, name, format):
        if format == 'pkl':
            return pd.read_pickle(name)
        elif format == 'csv':
            return audformat.utils.read_csv(name)
        else:
            self.error(f'unkown store format: {format}')

    def copy_flags(self, df_source, df_target):
        if hasattr(df_source, 'is_labeled'):
            df_target.is_labeled = df_source.is_labeled
        if hasattr(df_source, 'is_test'):
            df_target.is_test = df_source.is_test
        if hasattr(df_source, 'is_train'):
            df_target.is_train = df_source.is_train
        if hasattr(df_source, 'is_val'):
            df_target.is_val = df_source.is_val
        if hasattr(df_source, 'got_gender'):
            df_target.got_gender = df_source.got_gender
        if hasattr(df_source, 'got_age'):
            df_target.got_age = df_source.got_age
        if hasattr(df_source, 'got_speaker'):
            df_target.got_speaker = df_source.got_speaker


# augmenter.py
import pandas as pd
from nkululeko.util import Util 
from audiomentations import Compose, AddGaussianNoise, TimeStretch, PitchShift, Shift
import numpy as np
import audiofile
import os
from audformat.utils import map_file_path
import audeer

class Augmenter:
    """
        augmenting the train split
    """
    def __init__(self, df):
        self.df = df
        self.util = Util('augmenter')
        # Define a standard transformation that randomly add augmentations to files
        self.audioment = Compose([
            AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.015, p=0.5),
            TimeStretch(min_rate=0.8, max_rate=1.25, p=0.5),
            PitchShift(min_semitones=-4, max_semitones=4, p=0.5),
            Shift(min_fraction=-0.5, max_fraction=0.5, p=0.5),
        ])

    def changepath(self, fp, np):
#        parent = os.path.dirname(fp).split('/')[-1]
        fullpath = os.path.dirname(fp)
 #       newpath = f'{np}{parent}'
 #       audeer.mkdir(newpath)
        return fp.replace(fullpath, np)

    def augment(self, sample_selection):
        """
        augment the training files and return a dataframe with new files index.
        """
        files = self.df.index.get_level_values(0).values
        store = self.util.get_path('store')
        filepath = f'{store}augmentations/'
        audeer.mkdir(filepath)
        self.util.debug(f'augmenting {sample_selection} samples to {filepath}')
        newpath = ''
        for i, f in enumerate(files):
            signal, sr = audiofile.read(f)
            filename = os.path.basename(f)
            parent = os.path.dirname(f).split('/')[-1]
            sig_aug = self.audioment(samples = signal, sample_rate = sr)
            newpath = f'{filepath}/{parent}/'
            audeer.mkdir(newpath)
            audiofile.write(f'{newpath}{filename}', signal=sig_aug, sampling_rate=sr)
            if i%10==0:
                print(f'augmented {i} of {len(files)}')
        df_ret = self.df.copy()
        df_ret = df_ret.set_index(map_file_path(df_ret.index, lambda x: self.changepath(x, newpath)))
        aug_db_filename = self.util.config_val('DATA', 'augment_result', 'augment.csv')
        target = self.util.config_val('DATA', 'target', 'emotion')
        df_ret[target] = df_ret['class_label']
        df_ret = df_ret.drop(columns=['class_label'])
        df_ret.to_csv(aug_db_filename)
        return df_ret


# modelrunner.py

from nkululeko.util import Util
from nkululeko import glob_conf

class Modelrunner:
    """
    Class to model one run
    """
    def __init__(self, df_train, df_test, feats_train, feats_test, run):
        """Constructor setting up the dataframes
            Args:
                df_train: train dataframe
                df_test: test dataframe
                feats_train: train features
                feats_train: test features
        """
        self.df_train, self.df_test, self.feats_train, self.feats_test = df_train, df_test, feats_train, feats_test
        self.util = Util('modelrunner')
        self.run = run
        self.target = glob_conf.config['DATA']['target']
        # intialize a new model
        model_type = glob_conf.config['MODEL']['type']
        self._select_model(model_type)


    def do_epochs(self):
        # initialze results
        reports = []
        plot_epochs = self.util.config_val('PLOT', 'epochs', False)
        only_test = self.util.config_val('MODEL', 'only_test', False)
        # for all epochs
        for epoch in range(int(self.util.config_val('EXP', 'epochs', 1))):
            if only_test:
                self.model.load(self.run, epoch)
                self.util.debug(f'reusing model: {self.model.store_path}')
                self.model.reset_test(self.df_test, self.feats_test)
            else:
                self.model.set_id(self.run, epoch)
                self.model.train()
            report = self.model.predict()
            report.set_id(self.run, epoch)
            plot_name = self.util.get_plot_name()+f'_{self.run}_{epoch:03d}_cnf'
            reports.append(report)                
            self.util.debug(f'run: {self.run} epoch: {epoch}: result: '
                f'{reports[-1].get_result().get_test_result()}')
            if plot_epochs:
                self.util.debug(f'plotting conf matrix to {plot_name}')
                report.plot_confmatrix(plot_name, epoch)
            store_models = self.util.config_val('MODEL', 'save', False)
            plot_best_model = self.util.config_val('PLOT', 'best_model', False)
            if (store_models or plot_best_model) and (not only_test): # in any case the model needs to be stored to disk.
                self.model.store()
        if not plot_epochs:
            # Do at least one confusion matrix plot
            self.util.debug(f'plotting confusion matrix to {plot_name}')
            reports[-1].plot_confmatrix(plot_name, epoch)
        return reports

    def _select_model(self, model_type):
        if model_type=='svm':
            from nkululeko.model_svm import SVM_model
            self.model = SVM_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='svr':
            from nkululeko.model_svr import SVR_model
            self.model = SVR_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='xgb':
            from nkululeko.model_xgb import XGB_model
            self.model = XGB_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='xgr':
            from nkululeko.model_xgr import XGR_model
            self.model = XGR_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='bayes':
            from nkululeko.model_bayes import Bayes_model
            self.model = Bayes_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='gmm':
            from nkululeko.model_gmm import GMM_model
            self.model = GMM_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='knn':
            from nkululeko.model_knn import KNN_model
            self.model = KNN_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='knn_reg':
            from nkululeko.model_knn_reg import KNN_reg_model
            self.model = KNN_reg_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='tree':
            from nkululeko.model_tree import Tree_model
            self.model = Tree_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='tree_reg':
            from nkululeko.model_tree_reg import Tree_reg_model
            self.model = Tree_reg_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='cnn':
            from nkululeko.model_cnn import CNN_model
            from nkululeko.model_cnn import CNN_model
            self.model = CNN_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='mlp':
            from nkululeko.model_mlp import MLP_model
            self.model = MLP_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        elif model_type=='mlp_reg':
            from nkululeko.model_mlp_regression import MLP_Reg_model
            self.model = MLP_Reg_model(self.df_train, self.df_test, self.feats_train, self.feats_test)
        else:
            self.util.error(f'unknown model type: \'{model_type}\'')
        if self.util.exp_is_classification() and not self.model.is_classifier:
            self.util.error('Experiment type set to classification but model type is not a classifier')
        return self.model


import pandas as pd
from nkululeko.util import Util
import os

class FileChecker:
    def __init__(self, df):
        self.util = Util('filechecker')
        self.df = df.copy()
        self.util.copy_flags(df, self.df)
        check_vad = self.util.config_val('DATA', 'check_vad', False)
        if check_vad:
            self.util.debug(f'This may take a while downloading the VAD model')
            import torch
            torch.set_num_threads(1)
            self.vad_model, self.vad_utils = torch.hub.load(repo_or_dir='snakers4/silero-vad',
                              model='silero_vad',
                              force_reload=False,
                              onnx=False)

    def set_data(self, df):
        self.df = df.copy()
        self.util.copy_flags(df, self.df)

    def all_checks(self, data_name=''):
        self.check_size(data_name)
        return self.check_vad(data_name) 

    def check_size(self, data_name=''):
        """ limit number of samples
            the samples are selected randomly          
        """
        if data_name=='':
            min = self.util.config_val('DATA', 'check_size', False)
        else:
            min = self.util.config_val_data(data_name, 'check_size', False)
        if min:
            if min == "True":
                min = 1000 # 1000 bytes would be a reasonable minimal size for 16 kHz sr
            old_samples = self.df.shape[0]
            df = self.df.copy()
            for i in self.df.index:
                file = i[0]
                file_size = os.path.getsize(file)
                if file_size < int(min):
                    df = df.drop(i, axis=0)
            self.util.debug(f'{data_name}: checked for samples less than {min} bytes, reduced samples from {old_samples} to {df.shape[0]}')
            self.util.copy_flags(self.df, df) 
            self.df = df
            return df
        else:
            return self.df  
        

    def check_vad(self, data_name=''):
        """ limit number of samples
            the samples are selected randomly          
        """
        if data_name=='':
            check = self.util.config_val('DATA', 'check_vad', False)
        else:
            check = self.util.config_val_data(data_name, 'check_vad', False)
        if check:
            self.util.debug(f'{data_name}: checking for samples without speech.')
            SAMPLING_RATE = 16000
            (get_speech_timestamps,
            save_audio,
            read_audio,
            VADIterator,
            collect_chunks) = self.vad_utils
            old_samples = self.df.shape[0]
            df = self.df.copy()
            for i in self.df.index:
                file = i[0]
                wav = read_audio(file, sampling_rate=SAMPLING_RATE)
                speech_timestamps = get_speech_timestamps(wav, self.vad_model, sampling_rate=SAMPLING_RATE)
                if len(speech_timestamps)==0:
                    self.util.debug(f'{file}: no speech detected')
                    df = df.drop(i, axis=0)
            self.util.debug(f'{data_name}: checked for samples without speech, reduced samples from {old_samples} to {df.shape[0]}')
            self.util.copy_flags(self.df, df) 
            self.df = df
            return df
        else:
            return self.df  


        model, utils = torch.hub.load(repo_or_dir='snakers4/silero-vad',
                              model='silero_vad',
                              force_reload=False,
                              onnx=USE_ONNX)


import random
import os
import time
import numpy as np
from nkululeko.dataset import Dataset
from nkululeko.dataset_csv import Dataset_CSV
from nkululeko.dataset_ravdess import Ravdess
from nkululeko.filter_data import filter_min_dur
from nkululeko.runmanager import Runmanager
from nkululeko.test_predictor import Test_predictor
from nkululeko.feats_analyser import FeatureAnalyser
from nkululeko.util import Util
from nkululeko.feature_extractor import FeatureExtractor
from nkululeko.plots import Plots
from nkululeko.filter_data import DataFilter
from nkululeko.file_checker import FileChecker
import nkululeko.glob_conf as glob_conf
from nkululeko.demo_predictor import Demo_predictor
import ast # To convert strings to objects
import pandas as pd
from sklearn.preprocessing import LabelEncoder
from nkululeko.scaler import Scaler
import pickle
import audformat
import audeer 

class Experiment:
    """Main class specifying an experiment
    
    """
    

    def __init__(self, config_obj):
        """
        Parameters
        ----------
        config_obj : a config parser object that sets the experiment parameters and being set as a global object.
        """

        self.set_globals(config_obj)
        self.name = glob_conf.config['EXP']['name']
        self.root = glob_conf.config['EXP']['root']
        audeer.mkdir(self.root+self.name) # create the experiment directory
        self.util = Util('experiment')
        glob_conf.set_util(self.util)
        self.loso = self.util.config_val('MODEL', 'loso', False)
        self.logo = self.util.config_val('MODEL', 'logo', False)
        self.xfoldx = self.util.config_val('MODEL', 'k_fold_cross', False)
        self.start = time.process_time()

    def get_name(self):
        return self.util.get_exp_name()

    def set_globals(self, config_obj):
        """install a config object in the global space"""
        glob_conf.init_config(config_obj)

    def load_datasets(self):
        """Load all databases specified in the configuration and map the labels"""
        ds = ast.literal_eval(glob_conf.config['DATA']['databases'])
        self.datasets = {}
        self.got_speaker, self.got_gender, self.got_age = False, False, False
        for d in ds:
            # if d == 'ravdess':
            #     data = Ravdess()
            # else:
            ds_type = self.util.config_val_data(d, 'type', 'audformat')
            if ds_type == 'audformat':
                data = Dataset(d)
            elif ds_type == 'csv':
                data = Dataset_CSV(d)
            else:
                self.util.error(f'unknown data type: {ds_type}')
            data.load()
            data.prepare()
            if data.got_gender:
                self.got_gender = True
            if data.got_age:
                self.got_age = True
            if data.got_speaker:
                self.got_speaker = True
            self.datasets.update({d: data})
        self.target = self.util.config_val('DATA', 'target', 'emotion')

    def _import_csv(self, storage):
        # df = pd.read_csv(storage, header=0, index_col=[0,1,2])
        # df.index.set_levels(pd.to_timedelta(df.index.levels[1]), level=1)
        # df.index.set_levels(pd.to_timedelta(df.index.levels[2]), level=2)
        df = audformat.utils.read_csv(storage)
        df.is_labeled = True if self.target in df else False
        return df

    def fill_tests(self):
        """Only fill a new test set"""

        test_dbs = ast.literal_eval(glob_conf.config['DATA']['tests'])
        self.df_test = pd.DataFrame()
        start_fresh = eval(self.util.config_val('DATA', 'no_reuse', 'False'))
        store = self.util.get_path('store')
        storage_test = f'{store}extra_testdf.csv'
        if os.path.isfile(storage_test) \
            and not start_fresh:
            self.util.debug(f'reusing previously stored {storage_test}')
            self.df_test = self._import_csv(storage_test)
        else:
            for d in test_dbs:
                if d == 'ravdess':
                    data = Ravdess()
                else:
                    ds_type = self.util.config_val_data(d, 'type', 'audformat')
                    if ds_type == 'audformat':
                        data = Dataset(d)
                    elif ds_type == 'csv':
                        data = Dataset_CSV(d)
                    else:
                        self.util.error(f'unknown data type: {ds_type}')
                data.load()
                if data.got_gender:
                    self.got_gender = True
                if data.got_age:
                    self.got_age = True
                if data.got_speaker:
                    self.got_speaker = True
                data.prepare_labels()
                self.df_test = pd.concat([self.df_test, self.util.make_segmented_index(data.df)])
                self.df_test.is_labeled = data.is_labeled
            self.df_test.got_gender = self.got_gender
            self.df_test.got_speaker = self.got_speaker
            # self.util.set_config_val('FEATS', 'needs_features_extraction', 'True')
            # self.util.set_config_val('FEATS', 'no_reuse', 'True')
            self.df_test['class_labels'] = self.df_test[self.target]
            self.df_test[self.target] = self.label_encoder.transform(self.df_test[self.target])
            self.df_test.to_csv(storage_test)

    def fill_train_and_tests(self):
        """Set up train and development sets. The method should be specified in the config."""
        store = self.util.get_path('store')
        storage_test = f'{store}testdf.csv'
        storage_train = f'{store}traindf.csv'
        start_fresh = eval(self.util.config_val('DATA', 'no_reuse', 'False'))
        if os.path.isfile(storage_train) and os.path.isfile(storage_test) \
            and not start_fresh:
            self.util.debug(f'reusing previously stored {storage_test} and {storage_train}')
            self.df_test = self._import_csv(storage_test)
            self.df_train = self._import_csv(storage_train)
        else:
            self.df_train, self.df_test = pd.DataFrame(), pd.DataFrame()
            strategy = self.util.config_val('DATA', 'strategy', 'traintest')
            # some datasets against others in their entierty
            if strategy == 'cross_data':
                train_dbs = ast.literal_eval(glob_conf.config['DATA']['trains'])
                test_dbs = ast.literal_eval(glob_conf.config['DATA']['tests'])
                for dn in train_dbs:
                    d = self.datasets[dn]
                    d.prepare_labels()
                    self.df_train = self.df_train.append(self.util.make_segmented_index(d.df))
                    self.util.copy_flags(d, self.df_train)
                for dn in test_dbs:
                    d = self.datasets[dn]
                    d.prepare_labels()
                    self.df_test = self.df_test.append(self.util.make_segmented_index(d.df))
                    self.util.copy_flags(d, self.df_test)
            elif strategy == 'traintest':
                # default: train vs. test combined from all datasets
                for d in self.datasets.values():
                    d.split()
                    d.prepare_labels()
                    self.df_train = pd.concat([self.df_train, d.df_train])
                    self.util.copy_flags(d, self.df_train)
                    self.df_test = pd.concat([self.df_test, d.df_test])
                    self.util.copy_flags(d, self.df_test)
            else:
                self.util.error(f'unknown strategy: {strategy}')
            # save the file lists to disk for later reuse
            store = self.util.get_path('store')
            storage_test = f'{store}testdf.csv'
            storage_train = f'{store}traindf.csv'
            self.df_test.to_csv(storage_test)
            self.df_train.to_csv(storage_train)

        self.util.copy_flags(self, self.df_test)
        self.util.copy_flags(self, self.df_train)
        # Try data checks
        datachecker = FileChecker(self.df_train)
        self.df_train = datachecker.all_checks()
        datachecker.set_data(self.df_test)
        self.df_test = datachecker.all_checks()

        # Check for filters
        filter_sample_selection = self.util.config_val('DATA', 'filter.sample_selection', 'all')
        if filter_sample_selection=='all':
            datafilter = DataFilter(self.df_train)
            self.df_train = datafilter.all_filters()
            datafilter = DataFilter(self.df_test)
            self.df_test = datafilter.all_filters()
        elif filter_sample_selection=='train':
            datafilter = DataFilter(self.df_train)
            self.df_train = datafilter.all_filters()
        elif filter_sample_selection=='test':
            datafilter = DataFilter(self.df_test)
            self.df_test = datafilter.all_filters()
        else:
            self.util.error(f'unkown filter sample selection specifier {filter_sample_selection}, should be [all | train | test]')

        # encode the labels
        if self.util.exp_is_classification():
            datatype = self.util.config_val('DATA', 'type', 'dummy')
            if datatype == 'continuous':
                if self.df_test.is_labeled:
                    # remember the target in case they get labelencoded later
                    self.df_test['class_label'] = self.df_test[self.target]
                    test_cats = self.df_test['class_label'].unique()
                else:
                    # if there is no target, copy a dummy label
                    self.df_test = self._add_random_target(self.df_test)
                if self.df_train.is_labeled:
                    # remember the target in case they get labelencoded later
                    self.df_train['class_label'] = self.df_train[self.target]
                    train_cats = self.df_train['class_label'].unique()
            else:
                if self.df_test.is_labeled:
                    test_cats = self.df_test[self.target].unique()
                else:   
                    # if there is no target, copy a dummy label
                    self.df_test = self._add_random_target(self.df_test)
                train_cats = self.df_train[self.target].unique()
            if self.df_test.is_labeled:
                if type(test_cats) == np.ndarray:
                    self.util.debug(f'Categories test: {test_cats}')
                else:
                    self.util.debug(f'Categories test: {test_cats.to_list()}')
            if type(train_cats) == np.ndarray:
                self.util.debug(f'Categories train: {train_cats}')
            else:
                self.util.debug(f'Categories train: {train_cats.to_list()}')

            # encode the labels as numbers
            self.label_encoder = LabelEncoder()
            self.df_train[self.target] = self.label_encoder.fit_transform(self.df_train[self.target])
            self.df_test[self.target] = self.label_encoder.transform(self.df_test[self.target])
            glob_conf.set_label_encoder(self.label_encoder)
        if self.got_speaker:
            self.util.debug(f'{self.df_test.speaker.nunique()} speakers in test and {self.df_train.speaker.nunique()} speakers in train')

        target_factor = self.util.config_val('DATA', 'target_divide_by', False)
        if target_factor:
            self.df_test[self.target] = self.df_test[self.target] / float(target_factor)
            self.df_train[self.target] = self.df_train[self.target] / float(target_factor)
            if not self.util.exp_is_classification():
                self.df_test['class_label'] = self.df_test['class_label'] / float(target_factor)
                self.df_train['class_label'] = self.df_train['class_label'] / float(target_factor)

    def _add_random_target(self, df):
        labels = self.util.get_labels()
        a = [None]*len(df)
        for i in range(0, len(df)):
            a[i] = random.choice(labels)
        df[self.target] = a
        return df

    def plot_distribution(self):
        """Plot the distribution of samples and speaker per target class and biological sex"""
        plot = Plots()
        sample_selection = self.util.config_val('EXPL', 'sample_selection', 'all')
        if sample_selection=='all':
            df_labels = pd.concat([self.df_train, self.df_test])
            self.util.copy_flags(self.df_train, df_labels)
        elif sample_selection=='train':
            df_labels = self.df_train
            self.util.copy_flags(self.df_train, df_labels)
        elif sample_selection=='test':
            df_labels = self.df_test
            self.util.copy_flags(self.df_test, df_labels)
        else:
            self.util.error(f'unkown sample selection specifier {sample_selection}, should be [all | train | test]')

        plot.plot_distributions(df_labels)
        if self.got_speaker:
            plot.plot_distributions_speaker(df_labels)

    def extract_test_feats(self):
        self.feats_test = pd.DataFrame()
        feats_name = "_".join(ast.literal_eval(glob_conf.config['DATA']['tests']))
        self.feature_extractor = FeatureExtractor() 
        self.feature_extractor.set_data(self.df_test, feats_name, 'test')
        self.feats_test = self.feature_extractor.extract()
        self.util.debug(f'Test features shape:{self.feats_test.shape}')

    def extract_feats(self):
        """Extract the features for train and dev sets. 
        
        They will be stored on disk and need to be removed manually.
        
        The string FEATS.feats_type is read from the config, defaults to os. 
        
        """
        df_train, df_test = self.df_train, self.df_test
        feats_name = "_".join(ast.literal_eval(glob_conf.config['DATA']['databases']))
        self.feats_test, self.feats_train = pd.DataFrame(), pd.DataFrame()
        self.feature_extractor = FeatureExtractor() 
        self.feature_extractor.set_data(df_train, feats_name, 'train')
        self.feats_train =self.feature_extractor.extract()
        self.feature_extractor.set_data(df_test, feats_name, 'test')
        self.feats_test = self.feature_extractor.extract()
        self.util.debug(f'All features: train shape : {self.feats_train.shape}, test shape:{self.feats_test.shape}')

        self._check_scale()

    def augment(self):
        """
        Augment the selected samples
        """
        from nkululeko.augmenter import Augmenter
        sample_selection = self.util.config_val('DATA', 'augment', 'train')
        if sample_selection=='all':
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection=='train':
            df = self.df_train
        elif sample_selection=='test':
            df = self.df_test
        else:
            self.util.error(f'unknown augmentation selection specifier {sample_selection}, should be [all | train | test]')

        augmenter = Augmenter(df)
        augmenter.augment(sample_selection)

    def random_splice(self):
        """
        Random-splice the selected samples
        """
        from nkululeko.randomsplicer import Randomsplicer
        sample_selection = self.util.config_val('DATA', 'random_splice', 'train')
        if sample_selection=='all':
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection=='train':
            df = self.df_train
        elif sample_selection=='test':
            df = self.df_test
        else:
            self.util.error(f'unknown augmentation selection specifier {sample_selection}, should be [all | train | test]')
        randomsplicer = Randomsplicer(df)
        randomsplicer.run(sample_selection)


    def analyse_features(self, needs_feats):
        """
        Do a feature exploration
        
        """

        plot_feats = eval(self.util.config_val('EXPL', 'feature_distributions', 'False'))
        sample_selection = self.util.config_val('EXPL', 'sample_selection', 'all')

        if self.util.config_val('EXPL', 'value_counts', False):
            self.plot_distribution()
        if not needs_feats:
            return
        if sample_selection=='all':
            df_feats = pd.concat([self.feats_train, self.feats_test])
            df_labels = pd.concat([self.df_train, self.df_test])
        elif sample_selection=='train':
            df_feats = self.feats_train
            df_labels = self.df_train
        elif sample_selection=='test':
            df_feats = self.feats_test
            df_labels = self.df_test
        else:
            self.util.error(f'unknown sample selection specifier {sample_selection}, should be [all | train | test]')

        if plot_feats:
            feat_analyser = FeatureAnalyser(sample_selection, df_labels, df_feats)        
            feat_analyser.analyse()

        # check if a scatterplot should be done
        scatter_var = eval(self.util.config_val('EXPL', 'scatter', 'False'))
        if scatter_var:
            scatters = ast.literal_eval(glob_conf.config['EXPL']['scatter'])
            if self.util.exp_is_classification():
                plots = Plots()
                for scatter in scatters:
                    plots.scatter_plot(df_feats, df_labels['class_label'], scatter)
            else:
                 self.util.debug('can\'t do scatterplot if not classification')


        # check if a tsne should be plotted
        tsne = eval(self.util.config_val('EXPL', 'tsne', 'False'))
        if tsne: 
            if self.util.exp_is_classification():
                plots = Plots()
                all_feats =self.feats_train.append(self.feats_test)
                all_labels = self.df_train['class_label'].append(self.df_test['class_label'])
                plots.plotTsne(all_feats, all_labels, self.util.get_exp_name()+'_tsne')
            else:
                 self.util.debug('can\'t plot tsne if not classification')


    def _check_scale(self):
        scale = self.util.config_val('FEATS', 'scale', False)
        if scale: 
            self.scaler = Scaler(self.df_train, self.df_test, self.feats_train, self.feats_test, scale)
            self.feats_train, self.feats_test = self.scaler.scale()

    def init_runmanager(self):
        """Initialize the manager object for the runs."""
        self.runmgr = Runmanager(self.df_train, self.df_test, self.feats_train, self.feats_test)

    def run(self):
        """Do the runs."""
        self.runmgr.do_runs()

        # access the best results all runs
        self.reports = self.runmgr.best_results
        # try to save yourself
        save = self.util.config_val('EXP', 'save', False)
        if save: 
            # save the experiment for future use
            self.save(self.util.get_save_name())

        # self.__collect_reports()
        self.util.print_best_results(self.reports)

        # check if the test predictions should be saved to disk
        test_pred_file = self.util.config_val('EXP', 'save_test', False)
        if test_pred_file:
            self.predict_test_and_save(test_pred_file)

        # check if the majority voting for all speakers should be plotted
        conf_mat_per_speaker_function = self.util.config_val('PLOT', 'combine_per_speaker', False)
        if (conf_mat_per_speaker_function):
            self.plot_confmat_per_speaker(conf_mat_per_speaker_function)
        used_time = time.process_time() - self.start
        self.util.debug(f'Done, used {used_time:.3f} seconds')

        # check if a test set should be labeled by the model:
        label_data = self.util.config_val('DATA', 'label_data', False)
        label_result = self.util.config_val('DATA', 'label_result', False)
        if label_data and label_result:
            self.predict_test_and_save(label_result)

        return self.reports    

    def plot_confmat_per_speaker(self, function):
        if self.loso or self.logo or self.xfoldx:
            self.util.debug('plot combined speaker predictions not possible for cross validation')
            return
        best = self._get_best_report(self.reports)
        # if not best.is_classification:
        #     best.continuous_to_categorical()
        truths = best.truths
        preds = best.preds
        speakers = self.df_test.speaker.values
        print(f'{len(truths)} {len(preds)} {len(speakers) }')
        df = pd.DataFrame(data={'truth':truths, 'pred':preds, 'speaker':speakers})
        plot_name = 'result_combined_per_speaker'
        self.util.debug(f'plotting speaker combination ({function}) confusion matrix to {plot_name}')
        best.plot_per_speaker(df, plot_name, function)

    def _get_best_report(self, best_reports):
        return self.runmgr.get_best_result(best_reports)

    def print_best_model(self):
        self.runmgr.print_best_result_runs()

    def demo(self, file, is_list):
        model = self.runmgr.get_best_model()
        labelEncoder = None
        try: 
            labelEncoder = self.label_encoder
        except AttributeError:
            pass
        demo = Demo_predictor(model, file, is_list, self.feature_extractor, labelEncoder)
        demo.run_demo()

    def predict_test_and_save(self, result_name):
        model = self.runmgr.get_best_model()
        model.set_testdata(self.df_test, self.feats_test)
        test_predictor = Test_predictor(model, self.df_test, self.label_encoder, result_name)        
        test_predictor.predict_and_store()


    def load(self, filename):
        f = open(filename, 'rb')
        tmp_dict = pickle.load(f)
        f.close()          
        self.__dict__.update(tmp_dict) 

    def save(self, filename):
        try:
            f = open(filename, 'wb')
            pickle.dump(self.__dict__, f)
            f.close()
        except (AttributeError, TypeError) as error: 
            self.util.debug(f'Save experiment: Can\'t pickle local object: {error}')


# dataset_csv.py
import pandas as pd
import os
from nkululeko.dataset import Dataset
import os.path
import audformat.utils

class Dataset_CSV(Dataset):
    """ Class to represent datasets stored as a csv file"""

    def load(self):
        """Load the dataframe with files, speakers and task labels"""
        self.got_target, self.got_speaker, self.got_gender = False, False, False
        root = self.util.config_val_data(self.name, '', '')
        absolut_path = eval(self.util.config_val_data(self.name, 'absolute_path', True))
        self.util.debug(f'loading {self.name}')
#        df = pd.read_csv(root, index_col='file')       
        if not os.path.isabs(root):
            exp_root = self.util.config_val('EXP', 'root', '')
            root = exp_root + root
        df = audformat.utils.read_csv(root)       
        if not absolut_path:
            # add the root folder to the relative paths of the files 
            if audformat.index_type(df.index) == 'segmented':
                df = df.set_index(df.index.set_levels(df.index.levels[0].map(lambda x: os.path.dirname(root)+'/'+x), 0))
            else:
                df = df.set_index(df.index.to_series().apply(lambda x: os.path.dirname(root)+'/'+x)) 
        self.df = df
        self.db = None
        self.got_target = True
        self.is_labeled = self.got_target
        self.start_fresh = eval(self.util.config_val('DATA', 'no_reuse', 'False'))
        if self.is_labeled and not 'class_label' in self.df.columns:
            self.df['class_label'] = self.df[self.target]
        if 'gender' in df.columns:
            self.got_gender = True 
        if 'age' in df.columns:
            self.got_age = True 
        if 'speaker' in df.columns:
            self.got_speaker = True
            ns = df['speaker'].nunique()
            print(f'num of speakers: {ns}')

        self.util.debug(f'Loaded database {self.name} with {df.shape[0]} '\
            f'samples: got targets: {self.got_target}, got speakers: {self.got_speaker}, '\
            f'got sexes: {self.got_gender}')

    def prepare(self):
        super().prepare()


# xgrmodel.py

from xgboost.sklearn import XGBRegressor
from nkululeko.model import Model

class XGR_model(Model):
    """An XGBoost model"""

    is_classifier = False

    clf = XGBRegressor(use_label_encoder=False) # set up the regressor




# randomsplicer.py

"""
Code originally by Oliver Pauly

Based on an idea by Klaus Scherer

K. R. Scherer, “Randomized splicing: A note on a simple technique for masking speech content” 
Journal of Experimental Research in Personality, vol. 5, pp. 155–159, 1971.

Evaluated in:
F. Burkhardt, Anna Derington, Matthias Kahlau, Klaus Scherer, Florian Eyben and Björn Schuller: Masking Speech Contents by Random Splicing: is Emotional Expression Preserved?, Proc. ICASSP, 2023

"""

import pandas as pd
from nkululeko.util import Util 
import nkululeko.randomsplicing as rsp
import numpy as np
import audiofile as af
import os
from audformat.utils import map_file_path
import audeer

class Randomsplicer:
    """
        augmenting the train split
    """
    def __init__(self, df):
        self.df = df
        self.util = Util('randomsplicer')

    def changepath(self, fp, np):
        fullpath = os.path.dirname(fp)
        return fp.replace(fullpath, np)

    def run(self, sample_selection):
        """
        random splice the selected samples and return a dataframe with new files index.
            adjustable parameters:
            * p_reverse: probability of some samples to be in reverse order (default: 0.3)
            * top_db: top db level for silence to be recognized (default: 12)   
        """

        p_reverse=0.3
        top_db=12

        files = self.df.index.get_level_values(0).values
        store = self.util.get_path('store')
        filepath = f'{store}randomspliced/'
        audeer.mkdir(filepath)
        self.util.debug(f'random splicing {sample_selection} samples to {filepath}')
        newpath = ''
        for i, f in enumerate(files):
            signal, sr = af.read(f)
            filename = os.path.basename(f)
            parent = os.path.dirname(f).split('/')[-1]
            sig_new = rsp.random_splicing(
                signal, sr,
                p_reverse=p_reverse,
                top_db=top_db,
                )

            newpath = f'{filepath}/{parent}/'
            audeer.mkdir(newpath)
            af.write(f'{newpath}{filename}', signal=sig_new, sampling_rate=sr)
            if i%10==0:
                print(f'random spliced {i} of {len(files)}')
        df_ret = self.df.copy()
        df_ret = df_ret.set_index(map_file_path(df_ret.index, lambda x: self.changepath(x, newpath)))
        db_filename = self.util.config_val('DATA', 'random_splice_result', 'random_spliced.csv')
        target = self.util.config_val('DATA', 'target', 'emotion')
        df_ret[target] = df_ret['class_label']
        df_ret = df_ret.drop(columns=['class_label'])
        df_ret.to_csv(db_filename)
        return df_ret


# model_bayes.py

from sklearn.naive_bayes import GaussianNB
from nkululeko.model import Model

class Bayes_model(Model):
    is_classifier = True

    """An SVM model"""
    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        c = float(self.util.config_val('MODEL', 'C_val', '0.001'))
        self.clf = GaussianNB() # set up the classifier




# xgbmodel.py

from xgboost import XGBClassifier
from nkululeko.model import Model

class XGB_model(Model):
    """An XGBoost model"""

    is_classifier = True

    clf = XGBClassifier(use_label_encoder=False) # set up the classifier




# feats_import.py

from nkululeko.util import Util
from nkululeko.featureset import Featureset
import os
import pandas as pd
import audformat

class Importset(Featureset):
    """Class to import features that have been compiled elsewhere"""

    def __init__(self, name, data_df):
        super().__init__(name, data_df)

    def extract(self):
        """Import the features or load them from disk if present."""
        store = self.util.get_path('store')
        storage = f'{store}{self.name}.pkl'
        extract = eval(self.util.config_val('FEATS', 'needs_feature_extraction', False))
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        feat_import_file = self.util.config_val('FEATS', 'import_file', False)
        if not os.path.isfile(feat_import_file):
            self.util.warn(f'no import file: {feat_import_file}')
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug(f'importing features for {self.name}')
            # df = pd.read_csv(feat_import_file, sep=',', header=0, 
            #     index_col=['file', 'start', 'end'])
            df = audformat.utils.read_csv(feat_import_file)
            # scale features before use?
            # from sklearn.preprocessing import StandardScaler
            # scaler = StandardScaler()
            # scaled_features = scaler.fit_transform(df.values)
            # df = pd.DataFrame(scaled_features, index=df.index, columns=df.columns)
            # use only the rows from the data index
            #df = self.data_df.join(df).drop(columns=self.data_df.columns)
            df = df.loc[self.data_df.index]
            #df = pd.concat([self.data_df, df], axis=1, join="inner").drop(columns=self.data_df.columns)
            # in any case, store to disk for later use
            df.to_pickle(storage) 
            # and assign to be the "official" feature set
            self.df = df           
        else:
            self.util.debug('reusing imported features.')
            self.df = pd.read_pickle(storage)



# plots.py
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
from nkululeko.util import Util
import seaborn as sns
import numpy as np
import ast

class Plots():
    
    def __init__(self):
        """Initializing the util system"""
        self.util = Util('plots')
        self.format = self.util.config_val('PLOT', 'format', 'png')
        self.target = self.util.config_val('DATA', 'target', 'emotion')

    def plot_distributions_speaker(self, df):
        df_speakers = pd.DataFrame()
        pd.options.mode.chained_assignment = None  # default='warn'
        for s in df.speaker.unique():
            df_speaker = df[df.speaker == s]
            df_speaker['samplenum'] = df_speaker.shape[0]
            df_speakers = pd.concat([df_speakers, df_speaker.head(1)])
        # plot the distrubution of samples per speaker        
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        self.util.debug(f'plotting samples per speaker')
        if 'gender' in df_speakers:
            filename = f'samples_value_counts'
            ax = df_speakers.groupby('samplenum')['gender'].value_counts().unstack().plot(kind='bar', stacked=True, title=f'samples per speaker ({df_speakers.shape[0]})', rot=0)
            ax.set_ylabel(f'number of speakers')
            ax.set_xlabel('number of samples')
            fig = ax.figure
            plt.tight_layout()
            plt.savefig(f'{fig_dir}{filename}.{self.format}')
            plt.close(fig)
            fig.clear()
        else:
            filename = f'samples_value_counts'
            ax = df_speakers['samplenum'].value_counts().sort_values().plot(kind='bar', stacked=True, title=f'samples per speaker ({df_speakers.shape[0]})', rot=0)
            ax.set_ylabel(f'number of speakers')
            ax.set_xlabel('number of samples')
            fig = ax.figure
            plt.tight_layout()
            plt.savefig(f'{fig_dir}{filename}.{self.format}')
            plt.close(fig)
            fig.clear()

        self.plot_distributions(df_speakers, type='speakers')

    def plot_distributions(self, df, type='samples'):
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        attributes = ast.literal_eval(self.util.config_val('EXPL', 'value_counts', False))
        dist_type = self.util.config_val('EXPL', 'dist_type', 'hist')
        for att in attributes:
            if len(att) == 1:
                self.util.debug(f'plotting {att[0]}')
                filename = f'{self.target}-{att[0]}'
                if self.util.is_categorical(df[att[0]]):
                    ax = df.groupby('class_label')[att[0]].value_counts().unstack().plot(kind='bar', stacked=True, title=f'{type} {df.shape[0]}', rot=0)
                    ax.set_ylabel(f'number of {type}')
                    ax.set_xlabel(self.target)
                else:
                    if dist_type == 'hist':
                        ax = sns.histplot(df, x=att[0], hue='class_label', kde=True)
                        ax.set_title(f'{type} {df.shape[0]}')
                        ax.set_xlabel(f'value of {att[0]}')
                        ax.set_ylabel(f'number of {type}')
                    else:
                        ax = sns.displot(df, x=att[0], hue='class_label', kind="kde", fill=True)
                        ax.fig.suptitle(f'{type} {df.shape[0]}') 
                fig = ax.figure
                plt.tight_layout()
                plt.savefig(f'{fig_dir}{filename}_{type}.{self.format}')
                plt.close(fig)
                fig.clear()
            elif len(att) == 2:
                self.util.debug(f'plotting {att}')
                att1 = att[0]
                att2 = att[1]
                filename = f'{att[0]}-{att[1]}'
                filename = f'{self.target}-{filename}'
                if self.util.is_categorical(df[att1]):
                    ax = sns.scatterplot(data=df, x=self.target, y=att2, hue=att1)                                
                elif self.util.is_categorical(df[att2]):
                    ax = sns.scatterplot(data=df, x=self.target, y=att1, hue=att2)                
                else:
                    ax = sns.scatterplot(data=df, x=att1, y=att2, hue='class_label')
                fig = ax.figure
                ax.set_title(f'{type} {df.shape[0]}')
                plt.tight_layout()
                plt.savefig(f'{fig_dir}{filename}_{type}.{self.format}')   
                plt.close(fig)
                fig.clear()
            else:
                self.util.error(f'plot value counts: target {att} has more than 2 values')


    def describe_df(self, name, df, target, filename):
        """Make a stacked barplot of samples and speakers per sex and target values. speaker, gender and target columns must be present"""
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        sampl_num = df.shape[0]
        sex_col = 'gender'
        if target == 'gender':
            sex_col = 'class_label'
        if self.util.exp_is_classification() and target != 'gender':
            target = 'class_label'
        if df.is_labeled:
            if df.got_gender and df.got_speaker:
                spkr_num = df.speaker.nunique()
                female_smpl_num = df[df[sex_col]=='female'].shape[0]
                male_smpl_num = df[df[sex_col]=='male'].shape[0]
                self.util.debug(f'plotting {name}: # samples: {sampl_num} (f: {female_smpl_num}, m: '+\
                    f'{male_smpl_num}), # speakers: {spkr_num}')
                # fig, axes = plt.subplots(nrows=1, ncols=2)
                fig, axes = plt.subplots(nrows=1, ncols=1)
                # df.groupby(target)['gender'].value_counts().unstack().plot(kind='bar', stacked=True, ax=axes[0], \
                #     title=f'samples ({sampl_num})')
                df.groupby(target)['gender'].value_counts().unstack().plot(kind='bar', stacked=True, \
                    title=f'samples ({sampl_num})')
                # df.groupby(target)['speaker'].nunique().plot(kind='bar', ax=axes[1], title=f'speakers ({spkr_num})')
            else:
                self.util.debug(f'plotting {name}: # samples: {sampl_num}')
                fig, axes = plt.subplots(nrows=1, ncols=1)
                df[target].value_counts().plot(kind='bar', ax=axes, \
                    title=f'samples ({sampl_num})')
            plt.tight_layout()
            plt.savefig(f'{fig_dir}{filename}.{self.format}')
            fig.clear()
            plt.close(fig)

    def scatter_plot(self, feats, labels, dimred_type):
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        sample_selection = self.util.config_val('EXPL', 'sample_selection', 'all')
        filename = self.util.get_exp_name()+sample_selection+'_'+dimred_type
        filename = f'{fig_dir}{filename}.{self.format}'
        self.util.debug(f'computing {dimred_type}, this might take a while...')
        data = None
        if dimred_type == 'tsne':
            data = self.getTsne(feats)
        elif dimred_type == 'umap':
            import umap
            y_umap = umap.UMAP(n_neighbors=10, random_state=0,).fit_transform(feats.values)
            data = pd.DataFrame(
                y_umap,
                feats.index,
                columns=['Dim_1', 'Dim_2'],
            )
        elif dimred_type == 'pca':
            from sklearn.decomposition import PCA
            from sklearn.preprocessing import StandardScaler
            scaler = StandardScaler()
            pca = PCA(n_components=2)
            y_pca = pca.fit_transform(scaler.fit_transform(feats.values))
            data = pd.DataFrame(
                y_pca,
                feats.index,
                columns=['Dim_1', 'Dim_2'],
            )
        else:
            self.util.error(f'no such dimensionaity reduction functin: {dimred_type}')
        plot_data = np.vstack((data.T, labels)).T
        plot_df = pd.DataFrame(data=plot_data, columns=('Dim_1', 'Dim_2', 'label'))
        plt.tight_layout()
        ax = sns.FacetGrid(plot_df, hue='label', height=6).map(plt.scatter, 'Dim_1', 'Dim_2').add_legend()
        fig = ax.figure
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)

    def plotTsne(self, feats, labels, filename, perplexity=30, learning_rate=200):
        """Make a TSNE plot to see whether features are useful for classification"""
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        filename = f'{fig_dir}{filename}.{self.format}'
        self.util.debug(f'plotting tsne to {filename}, this might take a while...')
        model = TSNE(n_components=2, random_state=0, perplexity=perplexity, learning_rate=learning_rate)
        tsne_data = model.fit_transform(feats)
        tsne_data_labs = np.vstack((tsne_data.T, labels)).T
        tsne_df = pd.DataFrame(data=tsne_data_labs, columns=('Dim_1', 'Dim_2', 'label'))
        plt.tight_layout()
        ax = sns.FacetGrid(tsne_df, hue='label', height=6).map(plt.scatter, 'Dim_1', 'Dim_2').add_legend()
        fig = ax.figure
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)

    def getTsne(self, feats, perplexity=30, learning_rate=200):
        """Make a TSNE plot to see whether features are useful for classification"""
        model = TSNE(n_components=2, random_state=0, perplexity=perplexity, learning_rate=learning_rate)
        tsne_data = model.fit_transform(feats)
        return tsne_data

    def plot_feature(self, title, feature, label, df_labels, df_features):
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        filename = f'{fig_dir}feat_dist_{title}_{feature}.{self.format}'
        df_plot =  pd.DataFrame({label:df_labels[label], feature:df_features[feature]})
        ax = sns.violinplot(data=df_plot, x=label, y=feature)
        label = self.util.config_val('DATA', 'target', 'class_label')
        ax.set(title=f'{title} samples', xlabel = label)
        fig = ax.figure
        plt.tight_layout()
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)

    def plot_tree(self, model, features):
        from sklearn import tree
        ax = plt.gca()
        ax.figure.set_size_inches(100, 60)
#        tree.plot_tree(model, ax = ax)
        tree.plot_tree(model, feature_names=features.columns, ax = ax)
        plt.tight_layout()
        #print(ax)
        fig_dir = self.util.get_path('fig_dir')+'../' # one up because of the runs 
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val('PLOT', 'format', 'png')
        filename = f'{fig_dir}{exp_name}EXPL_tree-plot.{format}'
        fig = ax.figure
        fig.savefig(filename)
        fig.clear()
        plt.close(fig)


# feats_clap.py

from nkululeko.util import Util
from nkululeko.featureset import Featureset
import os
import pandas as pd
import os
import nkululeko.glob_conf as glob_conf
import laion_clap
import audiofile

class Clap(Featureset):
    """Class to extract laion's clap embeddings (https://github.com/LAION-AI/CLAP)"""

    def __init__(self, name, data_df):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df)
        self.device = self.util.config_val('MODEL', 'device', 'cpu')
        self.model_initialized = False



    def init_model(self):
        # load model
        self.util.debug('loading clap model...')
        self.model = laion_clap.CLAP_Module(enable_fusion=False)
        self.model.load_ckpt() # download the default pretrained checkpoint.
        print(f'loaded clap model')


    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path('store')
        store_format = self.util.config_val('FEATS', 'store_format', 'pkl')
        storage = f'{store}{self.name}.{store_format}'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug('extracting clap embeddings, this might take a while...')
            emb_series = pd.Series(index = self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(self.data_df.index.to_list()):
                signal, sampling_rate = audiofile.read(file, offset=start.total_seconds(), duration=(end-start).total_seconds(), always_2d=True)
                emb = self.get_embeddings(signal, sampling_rate)
                emb_series[idx] = emb
                if idx%10==0:
                    self.util.debug(f'Clap: {idx} of {length} done')
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index) 
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'false'
            except KeyError:
                pass
        else:
            self.util.debug('reusing extracted wav2vec2 embeddings')
            self.df = self.util.get_store(storage, store_format)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                print(nanrows)
                self.util.error(f'got nan: {self.df.shape} {self.df.isnull().sum().sum()}')


    def get_embeddings(self, signal, sampling_rate):
        tmp_audio_name = ['clap_audio_tmp.wav']
        audiofile.write(tmp_audio_name[0], signal, 48000)
        audio_embed = self.model.get_audio_embedding_from_filelist(x = tmp_audio_name, use_tensor=False)
        return audio_embed[0]
    
    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr)
        return feats


# model_mlp.py
from nkululeko.util import Util 
import nkululeko.glob_conf as glob_conf
from nkululeko.model import Model
from nkululeko.reporter import Reporter
import torch
import ast
import numpy as np
from sklearn.metrics import recall_score
from collections import OrderedDict
from nkululeko.loss_softf1loss import SoftF1Loss

class MLP_model(Model):
    """MLP = multi layer perceptron"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes"""
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.target = glob_conf.config['DATA']['target']
        labels = ast.literal_eval(glob_conf.config['DATA']['labels'])
        self.class_num = len(labels)
        # set up loss criterion
        criterion = self.util.config_val('MODEL', 'loss', 'cross')
        if criterion == 'cross':
            self.criterion = torch.nn.CrossEntropyLoss()
        elif criterion == 'f1':
            self.criterion = SoftF1Loss(num_classes=self.class_num, weight=None, epsilon=1e-7)
        else:
            self.util.error(f'unknown loss function: {criterion}')
        self.util.debug(f'using model with cross entropy loss function')
        # set up the model
        self.device = self.util.config_val('MODEL', 'device', 'cpu')
        layers_string = glob_conf.config['MODEL']['layers']
        self.util.debug(f'using layers {layers_string}')
        layers = ast.literal_eval(layers_string)
        # with dropout?
        drop = self.util.config_val('MODEL', 'drop', False)
        if drop:
            self.util.debug(f'init: training with dropout: {drop}')
        self.model = self.MLP(feats_train.shape[1], layers, self.class_num, drop).to(self.device)
        self.learning_rate = float(self.util.config_val('MODEL', 'learning_rate', 0.0001))
        # set up regularization
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate)
        # batch size
        self.batch_size = int(self.util.config_val('MODEL', 'batch_size', 8))
        # number of parallel processes
        self.num_workers = int(self.util.config_val('MODEL', 'num_workers', 5))
        if feats_train.isna().to_numpy().any():
            self.util.debug(f'Model, train: replacing {feats_train.isna().sum().sum()} NANs with 0')
            feats_train = feats_train.fillna(0)
        if feats_test.isna().to_numpy().any():
            self.util.debug(f'Model, test: replacing {feats_test.isna().sum().sum()} NANs with 0')
            feats_test = feats_test.fillna(0)
        # set up the data_loaders
        self.trainloader = self.get_loader(feats_train, df_train, True)
        self.testloader = self.get_loader(feats_test, df_test, False)

    def set_testdata(self, data_df, feats_df):
        self.testloader = self.get_loader(feats_df, data_df, False)

    def reset_test(self,  df_test, feats_test):
        self.testloader = self.get_loader(feats_test, df_test, False)

    def train(self):            
        self.model.train()
        losses = []
        for features, labels in self.trainloader:
            logits = self.model(features.to(self.device))
            loss = self.criterion(logits, labels.to(self.device, dtype=torch.int64))
            losses.append(loss.item())
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
        self.loss = (np.asarray(losses)).mean()

    def predict(self):        
        _, truths, predictions = self.evaluate_model(self.model, self.testloader, self.device)
        uar, _, _ = self.evaluate_model(self.model, self.trainloader, self.device)
        report = Reporter(truths, predictions, self.run, self.epoch)
        try:
            report.result.loss = self.loss
        except AttributeError: # if the model was loaded from disk the loss is unknown
            pass 
        report.result.train = uar
        return report

    def get_predictions(self):
        _, truths, predictions = self.evaluate_model(self.model, self.testloader, self.device)
        return predictions.numpy()

    def get_loader(self, df_x, df_y, shuffle):
        data=[]
        for i in range(len(df_x)):
            data.append([df_x.values[i], df_y[self.target][i]])
        return torch.utils.data.DataLoader(data, shuffle=shuffle, batch_size=self.batch_size)

    class MLP(torch.nn.Module):
        def __init__(self, i, layers, o, drop):
            super().__init__()
            sorted_layers = sorted(layers.items(), key=lambda x: x[1])
            layers = OrderedDict()
            layers['0'] = torch.nn.Linear(i, sorted_layers[0][1])
            layers['0_r'] = torch.nn.ReLU()
            for i in range(0, len(sorted_layers)-1):         
                layers[str(i+1)] = torch.nn.Linear(sorted_layers[i][1], sorted_layers[i+1][1])
                if drop:
                    layers[str(i)+'_d'] = torch.nn.Dropout(float(drop))
                layers[str(i)+'_r'] = torch.nn.ReLU()
            layers[str(len(sorted_layers)+1)] = torch.nn.Linear(sorted_layers[-1][1], o)
            self.linear = torch.nn.Sequential(layers)
        def forward(self, x):
            # x: (batch_size, channels, samples)
            x = x.squeeze(dim=1).float()
            return self.linear(x)


    def evaluate_model(self, model, loader, device):
        logits = torch.zeros(len(loader.dataset), self.class_num)
        targets = torch.zeros(len(loader.dataset))
        model.eval()
        with torch.no_grad():
            for index, (features, labels) in enumerate(loader):
                start_index = index * loader.batch_size
                end_index = (index + 1) * loader.batch_size
                if end_index > len(loader.dataset):
                    end_index = len(loader.dataset)
                logits[start_index:end_index, :] = model(features.to(device))
                targets[start_index:end_index] = labels

        predictions = logits.argmax(dim=1)
        uar = recall_score(targets.numpy(), predictions.numpy(), average='macro')
        return uar, targets, predictions

    def predict_sample(self, features):
        """Predict one sample"""
        with torch.no_grad():
            logits = self.model(torch.from_numpy(features).to(self.device))
        a = logits.numpy()
        res = {}
        for i in range(len(a[0])):
            res[i] = a[0][i]
        return res

    def store(self):
        torch.save(self.model.state_dict(), self.store_path)
        
    def load(self, run, epoch):
        self.set_id(run, epoch)
        dir = self.util.get_path('model_dir')
        # name = f'{self.util.get_exp_name()}_{run}_{epoch:03d}.model'
        name = f'{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model'
        self.device = self.util.config_val('MODEL', 'device', 'cpu')
        layers = ast.literal_eval(glob_conf.config['MODEL']['layers'])
        self.store_path = dir+name
        drop = self.util.config_val('MODEL', 'drop', False)
        if drop:
            self.util.debug(f'loading: dropout set to: {drop}')
        self.model = self.MLP(self.feats_train.shape[1], layers, self.class_num, drop).to(self.device)
        self.model.load_state_dict(torch.load(self.store_path))
        self.model.eval()
    
    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, 'rb') as handle:
            self.device = self.util.config_val('MODEL', 'device', 'cpu')
            layers = ast.literal_eval(glob_conf.config['MODEL']['layers'])
            self.store_path = path
            drop = self.util.config_val('MODEL', 'drop', False)
            if drop:
                self.util.debug(f'dropout set to: {drop}')
            self.model = self.MLP(self.feats_train.shape[1], layers, self.class_num, drop).to(self.device)
            self.model.load_state_dict(torch.load(self.store_path))
            self.model.eval()